/*
    Simon 68k v1.0
    Copyright (C) 2001 John David Ratliff
    Questions or Comments: 
    	E-Mail: jdratlif@cs.indiana.edu
    	URL:	http://www.technoplaza.net 
    	
    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

// C Source File
// Created 6/2/01; 11:46:07 PM

#define OPTIMIZE_ROM_CALLS    // Use ROM Call Optimization

#define SAVE_SCREEN           // Save/Restore LCD Contents

#include "simon.h"

#include <stdio.h>
#include <stdlib.h>
#include <system.h>
#include <sprites.h>
#include <graph.h>
#include <kbd.h>
#include <compat.h>
#include <dialogs.h>
#include <alloc.h>
#include <string.h>

#ifdef _TI89_
short _ti89;                  // Produce .89Z File
#endif

#ifdef _TI92_
short _ti92plus;              // Produce .9XZ File
#endif

void printTile(short int tile) {
	short int x, y;
	
	switch (tile) {
		case LEFT: x = LEFTX; y = LEFTY; break;
		case RIGHT: x = RIGHTX; y = RIGHTY; break;
		case DOWN: x = DOWNX; y = DOWNY; break;
		case UP: x = UPX; y = UPY; break;
		default: tile = LEFT; x = 0; y = 0; break;
	}
	
	Sprite32(x,y,SPRITE_HEIGHT,patterns[tile],LCD_MEM,SPRT_XOR);
	
	// pause to display sprite
	OSTimerRestart(USER_TIMER);
	while (!OSTimerExpired(USER_TIMER));
	
	Sprite32(x,y,SPRITE_HEIGHT,patterns[tile],LCD_MEM,SPRT_XOR);
}

void playSimonPattern(short int *pattern, short int len) {
	short int loop;
	
	for (loop = 0; loop < len; loop++) {
		printTile(pattern[loop]);
		
		OSTimerRestart(USER_TIMER);
		while (!OSTimerExpired(USER_TIMER));
	}
}

void getUserPattern(short int *pattern, short int maxlen) {
	short int done = 0, loop = 0;
	unsigned short int key = 0;
	void *kbd = kbd_queue();
	
	do {
		while (OSdequeue(&key,kbd))
			idle();
		
		if ((signed short int)key == KEY_LEFT) {
			printTile(LEFT);
			pattern[loop] = LEFT;
		} else if ((signed short int)key == KEY_RIGHT) {
			printTile(RIGHT);
			pattern[loop] = RIGHT;
		} else if ((signed short int)key == KEY_DOWN) {
			printTile(DOWN);
			pattern[loop] = DOWN;
		} else if ((signed short int)key == KEY_UP) {
			printTile(UP);
			pattern[loop] = UP;
		} else if ((signed short int)key == KEY_ESC) {
			done = 1;
			pattern[0] = key;
		} else {
			--loop;
		}
	} while (!done && ++loop < maxlen);
}

void displayLevel(short int level, short int patterns) {
	short int x = 72, y = 40, height = 16;
	
	if (level < (patterns - 1)) {
		Sprite16(x,y,height,levels[level+1],LCD_MEM,SPRT_XOR);
	}
	
	Sprite16(x,y,height,levels[level],LCD_MEM,SPRT_XOR);
}

void drawLogo(void) {
	short int offset=0,x,y,width=16,height=31,vcount=1,hcount=10;
	
	for (y = 0; y < vcount; y++) {
		for (x = 0; x < hcount; x++) { 
			Sprite16(x*width,y*height,height,&logo[offset],LCD_MEM,SPRT_XOR);
			offset += 31;
		}
	}
}

inline void drawBorder(void) {
	DrawLine(X1,Y1,X2,Y2,A_NORMAL);
	DrawLine(X3,Y3,X4,Y4,A_NORMAL);
	DrawLine(X5,Y5,X6,Y6,A_NORMAL);
	DrawLine(X7,Y7,X8,Y8,A_NORMAL);
}

inline void drawMenu(void) {
	DrawStr(15,40,(char *)intro[0],A_XOR);
	DrawStr(23,50,(char *)intro[1],A_XOR);
	DrawStr(20,65,(char *)intro[2],A_XOR);
}

inline void drawCopyright(void) {
	FontSetSys(F_4x6);
	DrawStr(15,85,(char *)intro[3],A_XOR);
	DrawStr(25,92,(char *)intro[4],A_XOR);
	FontSetSys(F_6x8);
}

inline void drawOptionsMenu(void) {
	DrawStr(5,40,(char *)options[SPEED],A_XOR);
	DrawStr(5,50,(char *)options[DIFFICULTY],A_XOR);
	DrawStr(5,60,(char *)options[EXIT],A_XOR);
}

inline void drawSpeed(short int speed) {
	DrawStr(75,40,(char *)speeds[speed],A_XOR);
}

inline void drawDifficultyLevel(short int difficulty) {
	DrawStr(75,50,(char *)difficulties[difficulty],A_XOR);
}

inline void dissipate(void) {
	short int wait1, wait2;
	
	for (wait1 = 0; wait1 < STATUS_WAIT; wait1++);
		for (wait2 = 0; wait2 < STATUS_WAIT; wait2++);
}

short int doIntro(short int *speed, short int *difficulty) {
	short int key = 0, options = 0;
	
	// allow time for status line to dissipate
	dissipate();
	
	// clear the screen
	ClrScr();
	
	// draw the border
	drawBorder();
	
	// draw the logo
	drawLogo();
	
	// draw the main menu
	drawMenu();
	
	// draw the copyright
	drawCopyright();
	
	do {
		key = ngetchx();
		
		if (options) {
			if (key == KEY_F5) {
				drawOptionsMenu();
				drawMenu();
				
				drawSpeed(*speed);
				drawDifficultyLevel(*difficulty);
				
				options = 0;
			} else if (key == KEY_F2) {
				drawSpeed(*speed);
				
				if (*speed < 4) {
					++(*speed);
				} else {
					*speed = 0;
				}
				
				drawSpeed(*speed);
			} else if (key == KEY_F3) {
				drawDifficultyLevel(*difficulty);
				
				if (*difficulty < 4) {
					++(*difficulty);
				} else {
					*difficulty = 0;
				}
				
				drawDifficultyLevel(*difficulty);
			}
		} else if (key == KEY_F1) {
			options = 1;
			
			drawMenu();
			drawOptionsMenu();
			
			drawSpeed(*speed);
			drawDifficultyLevel(*difficulty);
		}
		
		// loop in idle state to save power
		idle();
	} while ((key != KEY_ESC && key != KEY_ENTER) || options == 1);
	
	return ((key == KEY_ENTER) ? 0 : 1);
}

inline void drawLetters(void) {
	Sprite16(2,2,16,letters,LCD_MEM,SPRT_XOR);
	Sprite16(2,18,16,letters+16,LCD_MEM,SPRT_XOR);
	Sprite16(2,34,16,letters+32,LCD_MEM,SPRT_XOR);
	Sprite16(2,50,16,letters+48,LCD_MEM,SPRT_XOR);
	Sprite16(2,66,16,letters+64,LCD_MEM,SPRT_XOR);
}

inline void displayScore(unsigned long int score) {
	printf_xy(3,90,"%05lu",score);
}

void doHiscoreBoard(unsigned long int score) {
	short int loop, pos = -1;
	HANDLE dlg = H_NULL;
	const char *title = "Memory Error", *error = "No Free Memory";
	char str[45],name[10];
	
	// erase the name variable
	name[0] = 0;
	
	for (loop = 4; loop >= 0; loop--) {
		if (score > hiscores[loop]) {
			// new hiscore!
			pos = loop;
		}
	}
	
	if (pos != -1) {
		if ((dlg = DialogNewSimple(HISCORE_WIDTH,HISCORE_HEIGHT)) == H_NULL) {
			DlgMessage((char *)title,(char *)error,BT_OK,BT_NONE);
		} else {
			DialogAddTitle(dlg,"New Hiscore!",BT_OK,BT_NONE);
			
			sprintf(str,"You earned the #%hd hiscore position!",pos+1);
			DialogAddText(dlg,5,15,str);
			
			DialogAddRequest(dlg,5,25,"Your Name:",0,9,11);
			
			while (DialogDo(dlg,CENTER,CENTER,name,NULL) != KEY_ENTER);
			
			HeapFree(dlg);
			
			// move the hiscore list down
			if (pos < 4) {
				for (loop = 4; loop > pos; loop--) {
					hiscores[loop] = hiscores[loop - 1];
					hiscorenames[loop][0] = 0;
					strcpy(hiscorenames[loop],hiscorenames[loop - 1]);
				}
			}
			
			// assign the new hiscore
			hiscores[pos] = score;
			
			// erase the old hiscore name
			hiscorenames[pos][0] = 0;
			
			// fill in the new name
			strcpy(hiscorenames[pos],name);
		}
	}
	
	// display the hiscore board
	
	// clear the screen
	ClrScr();
	
	// draw the screen borders
	drawBorder();
	
	// draw the game logo
	drawLogo();
	
	pos = FontGetSys();
	FontSetSys(F_8x10);
	DrawStr(25,35,"Hiscore Board",A_NORMAL);
	FontSetSys(pos);
	
	for (loop = 0; loop < 5; loop++) {
		printf_xy(20,50+loop*10,"#%hd %-9s %lu",loop+1,hiscorenames[loop],hiscores[loop]);
	}
	
	ngetchx();
}

short int getWinnersBonus(short int speed, short int difficulty) {
	short int bonus = 0;
	
	switch (difficulty) {
		case VERYEASY: bonus += 250; break;
		case EASY: bonus += 1000; break;
		case AVG: bonus += 4000; break;
		case HARD: bonus += 10000; break;
		case VERYHARD: bonus += 20000; break;
	}
		
	switch (speed) {
		case VERYSLOW: bonus += 250; break;
		case SLOW: bonus += 1000; break;
		case AVERAGE: bonus += 3000; break;
		case FAST: bonus += 6000; break;
		case VERYFAST: bonus += 12000; break;
	}
	
	return bonus;
}

void play(short int speed, short int difficulty) {
	short int user[16], simon[16], pos = 0, lost = 0, loop, patterns = difficultyLevels[difficulty];
	short int bonus = getWinnersBonus(speed,difficulty), done = 0, set = 0;
	unsigned long int score = 0;
	
	// erase the copyright info
	drawCopyright();

	// draw the intro string
	DrawStr(2,85,"Get Ready to Play Simon...",A_XOR);
	
	// wait before displaying first piece...
	for (loop = 0; loop < 5; loop ++) {
		OSTimerRestart(USER_TIMER);
		while (!OSTimerExpired(USER_TIMER));
	}
	
	// clear the screen
	ClrScr();	
	
	// draw the borders
	drawBorder();
	
	// display game title block letters
	drawLetters();
	
	// draw initial score
	displayScore(score);
	
	while (!done) {
		while (pos < patterns && !lost) {
			displayLevel((patterns - pos) - 1, patterns);
		
			simon[pos] = rand() % 4;
			user[pos] = 4;
		
			playSimonPattern(simon,++pos);
			getUserPattern(user,pos);
		
			for (loop = (pos - 1); loop >= 0; loop--) {
				if (user[loop] != simon[loop]) {
					lost = 1;
				}
			}
		
			if (!lost) {
				// increment and redraw score
				score += (7 * pos * (speed + 1) * (difficulty + 1));
				displayScore(score);
			}

			// wait before switching to new Simon pattern
			OSTimerRestart(USER_TIMER);
			while (!OSTimerExpired(USER_TIMER));
		};
		
		if (lost) {
			done = 1;
			
			if (user[0] == KEY_ESC) {
				DlgMessage((char *)title,"Leaving so soon?",BT_OK,BT_NONE);
			} else {
				DlgMessage((char *)title,"Too bad. Try Again.",BT_OK,BT_NONE);
			}
		} else {
			++set;
			
			// add the winner's bonus
			score += bonus;
			
			// redraw the score
			displayScore(score);
			
			if (set < 5) {
				DlgMessage((char *)title,"Get Ready for the Next Set!",BT_OK,BT_NONE);
				
				// reset the position
				pos = 0;
			
				// erase the old pattern indicator
				Sprite16(72,40,16,levels[0],LCD_MEM,SPRT_XOR);
			} else {
				// give the big winner bonus
				score += (((unsigned long int)bonus) * 10);
				displayScore(score);				
				
				DlgMessage((char *)title,"Great Job! You won all 5 sets!",BT_OK,BT_NONE);
				
				done = 1;
			}
		}
		
		// wait for status line to dissipate
		dissipate();
	};
	
	// show the hi score board, and see if we got a new hiscore
	doHiscoreBoard(score);
}

// Main Function
void _main(void)
{
	short int speed = 2, difficulty = 2, font = FontGetSys();;
	
	// seed the random number generator
	randomize();
	
	// setup the standard font
	FontSetSys(F_6x8);
	
	while (doIntro(&speed,&difficulty) == 0) {
		// set display timer
		OSFreeTimer(USER_TIMER);
		OSRegisterTimer(USER_TIMER,speedTimers[speed]);

		play(speed,difficulty);
	};
	
	// reset to the entry font
	FontSetSys(font);
}
